// license: AGPLv3 or later
// copyright 2023 zamfofex

#include <mimimi/text.h>
#include <mimimi/engines.h>
#include <mimimi/font.h>

#include <mimimi/chapter-I/dialogues.h>

static void mimimi_chapter_I_box(struct mimimi_image *image, unsigned char stroke, unsigned char fill)
{
	int width = image->width;
	int height = image->height;
	
	unsigned char table[] = {0, stroke, fill};
	static unsigned char colors[10][10] =
	{
		{0,0,0,0,0,0,0,0,0,1},
		{0,0,0,0,0,0,0,1,1,1},
		{0,0,0,0,0,1,1,1,1,1},
		{0,0,0,0,1,1,1,1,2,2},
		{0,0,0,1,1,1,2,2,2,2},
		{0,0,1,1,1,2,2,2,2,2},
		{0,0,1,1,2,2,2,2,2,2},
		{0,1,1,1,2,2,2,2,2,2},
		{0,1,1,2,2,2,2,2,2,2},
		{1,1,1,2,2,2,2,2,2,2},
	};
	
	for (int y = 0 ; y < height ; y++)
	for (int x = 0 ; x < width ; x++)
		image->colors[x + y * width] = fill;
	
	for (int x = 0 ; x < width ; x++)
	{
		image->colors[x + 0 * width] = stroke;
		image->colors[x + 1 * width] = stroke;
		image->colors[x + (height - 1) * width] = stroke;
		image->colors[x + (height - 2) * width] = stroke;
	}
	
	for (int y = 0 ; y < height ; y++)
	{
		image->colors[0 + y * width] = stroke;
		image->colors[1 + y * width] = stroke;
		image->colors[(width - 1) + y * width] = stroke;
		image->colors[(width - 2) + y * width] = stroke;
	}
	
	for (int y0 = 0 ; y0 < 10 ; y0++)
	for (int x0 = 0 ; x0 < 10 ; x0++)
	{
		int x1 = width - 1 - x0;
		int y1 = height - 1 - y0;
		
		unsigned char color = table[colors[y0][x0]];
		
		image->colors[x0 + y0 * width] = color;
		image->colors[x1 + y0 * width] = color;
		image->colors[x0 + y1 * width] = color;
		image->colors[x1 + y1 * width] = color;
	}
}

static void mimimi_chapter_I_portrait(struct mimimi_image *image, struct mimimi_model *model, unsigned char background, struct mimimi_allocator *allocator)
{
	int width = image->width;
	int height = image->height;
	
	struct mimimi_pose_layer layer = {0, 42, 0};
	struct mimimi_pose pose = {width / 2, height / 2 + 16, 1, &layer};
	
	mimimi_pose(image, model, &pose, allocator);
	
	for (int i = 0 ; i < width * height ; i++)
	{
		if (image->colors[i] == 0)
			image->colors[i] = background;
	}
	
	static unsigned char mask[6][6] =
	{
		{0,0,0,0,0,0},
		{0,0,0,0,1,1},
		{0,0,0,1,1,1},
		{0,0,1,1,1,1},
		{0,1,1,1,1,1},
		{0,1,1,1,1,1},
	};
	
	for (int y0 = 0 ; y0 < 6 ; y0++)
	for (int x0 = 0 ; x0 < 6 ; x0++)
	{
		if (mask[y0][x0] != 0) continue;
		
		int x1 = width - 1 - x0;
		int y1 = height - 1 - y0;
		
		image->colors[x0 + y0 * width] = 0;
		image->colors[x1 + y0 * width] = 0;
		image->colors[x0 + y1 * width] = 0;
		image->colors[x1 + y1 * width] = 0;
	}
}

static void mimimi_chapter_I_stamp(struct mimimi_image *target, int x0, int y0, struct mimimi_image *source)
{
	for (int y1 = 0 ; y1 < source->height ; y1++)
	for (int x1 = 0 ; x1 < source->width ; x1++)
	{
		int x = x0 + x1;
		int y = y0 + y1;
		
		if (x < 0) continue;
		if (y < 0) continue;
		
		if (x >= target->width) break;
		if (y >= target->height) return;
		
		unsigned char color = source->colors[x1 + y1 * source->width];
		if (color == 0) continue;
		
		target->colors[x + y * target->width] = color;
	}
}

static struct mimimi_image *mimimi_chapter_I_dialogue_box(struct mimimi_image *box, struct mimimi_model *model, unsigned char *name, int width, unsigned char *paragraph, struct mimimi_allocator *allocator)
{
	int line_count = 0;
	for (unsigned char *text = paragraph ; *text != 0 ; text = mimimi_skip_line(text, width - 12, mimimi_font))
		line_count++;
	
	int height = 30 + 14 * line_count;
	
	box->width = width;
	box->height = height;
	box->colors = mimimi_allocate(allocator, width * height);
	mimimi_chapter_I_box(box, 4, 13);
	
	struct mimimi_allocator *portrait_allocator = mimimi_temporary_allocator(allocator);
	
	struct mimimi_image *portrait = mimimi_allocate(portrait_allocator, sizeof *portrait);
	portrait->width = 18;
	portrait->height = 18;
	portrait->colors = mimimi_allocate(portrait_allocator, portrait->width * portrait->height);
	mimimi_chapter_I_portrait(portrait, model, 8, portrait_allocator);
	mimimi_chapter_I_stamp(box, 6, 6, portrait);
	
	mimimi_finish_temporary_allocator(portrait_allocator);
	
	int n = mimimi_draw_line(box, 28, 17, name, name, mimimi_count_paragraph(name), mimimi_font, 4);
	for (int i = 28 ; i < n - 1 ; i++)
		box->colors[i + 18 * box->width] = 4;
	
	mimimi_draw_text(box, 6, 34, width - 12, 14, paragraph, mimimi_font, 4);
	return box;
}

struct mimimi_chapter_I_strip_dialogue
{
	int count;
	void *texture;
	int *notches;
};

struct mimimi_chapter_I_strip_dialogue *mimimi_chapter_I_strip_dialogue(struct mimimi_chapter_I_dialogue *dialogue, struct mimimi_engine *engine, struct mimimi_allocator *allocator)
{
	int *notches = mimimi_allocate(allocator, dialogue->count * sizeof *notches);
	struct mimimi_image *images = mimimi_allocate(allocator, dialogue->count * sizeof *images);
	int width = 128;
	int height = 0;
	
	for (int i = 0 ; i < dialogue->count ; i++)
	{
		struct mimimi_chapter_I_paragraph *paragraph = dialogue->paragraphs + i;
		mimimi_chapter_I_dialogue_box(images + i, paragraph->model, paragraph->name, width, paragraph->text, allocator);
		height += images[i].height + 8;
		notches[i] = height;
	}
	
	struct mimimi_image *image = mimimi_allocate(allocator, sizeof *image);
	image->width = width;
	image->height = height;
	image->colors = mimimi_allocate(allocator, width * height);
	for (int i = 0 ; i < width * height ; i++)
		image->colors[i] = 0;
	
	mimimi_chapter_I_stamp(image, 0, 0, images);
	for (int i = 1 ; i < dialogue->count ; i++)
		mimimi_chapter_I_stamp(image, 0, notches[i - 1], images + i);
	
	void *texture = (*engine->texture)(engine->data, image);
	
	struct mimimi_chapter_I_strip_dialogue *strip_dialogue = mimimi_allocate(allocator, sizeof *strip_dialogue);
	strip_dialogue->count = dialogue->count;
	strip_dialogue->notches = notches;
	strip_dialogue->texture = texture;
	
	return strip_dialogue;
}

void mimimi_chapter_I_show_strip_dialogue(struct mimimi_behavior *behavior, struct mimimi_engine *engine, struct mimimi_chapter *chapter, struct mimimi_chapter_I_dialogue *dialogue, struct mimimi_allocator *allocator)
{
	struct mimimi_allocator *temporary_allocator = mimimi_temporary_allocator(allocator);
	struct mimimi_chapter_I_strip_dialogue *strip_dialogue = mimimi_chapter_I_strip_dialogue(dialogue, engine, temporary_allocator);
	
	int i = 0;
	int y = 0;
	int held = 1;
	
	while (i < strip_dialogue->count)
	{
		(*engine->stamp)(engine->data, 8, engine->size->height - y, strip_dialogue->texture);
		
		int distance = strip_dialogue->notches[i] - y;
		if (distance > 0) y += (distance - 1) / 8 + 1;
		
		if (chapter->left != 0 || chapter->right != 0)
		{
			if (held == 0) i++;
			held = 1;
		}
		else
		{
			held = 0;
		}
		
		(*behavior->behave)(behavior->data);
	}
	
	for (int i = 0 ; i < 32 ; i++)
	{
		(*engine->stamp)(engine->data, 8 - (128 + 8) * i / 32, engine->size->height - strip_dialogue->notches[strip_dialogue->count - 1], strip_dialogue->texture);
		(*behavior->behave)(behavior->data);
	}
	
	mimimi_finish_temporary_allocator(temporary_allocator);
	(*engine->invalidate)(engine->data, strip_dialogue->texture);
}

struct mimimi_chapter_I_toast_dialogue
{
	int count;
	void **textures;
	int *delays;
	int *heights;
};

struct mimimi_chapter_I_toast_dialogue *mimimi_chapter_I_toast_dialogue(struct mimimi_chapter_I_dialogue *dialogue, struct mimimi_engine *engine, struct mimimi_allocator *allocator)
{
	int *delays = mimimi_allocate(allocator, dialogue->count * sizeof *delays);
	void **textures = mimimi_allocate(allocator, dialogue->count * sizeof *textures);
	int *heights = mimimi_allocate(allocator, dialogue->count * sizeof *heights);
	struct mimimi_image *images = mimimi_allocate(allocator, dialogue->count * sizeof *images);
	
	int width = 128;
	
	for (int i = 0 ; i < dialogue->count ; i++)
	{
		struct mimimi_chapter_I_paragraph *paragraph = dialogue->paragraphs + i;
		
		mimimi_chapter_I_dialogue_box(images + i, paragraph->model, paragraph->name, width, paragraph->text, allocator);
		heights[i] = images[i].height + 8;
		textures[i] = (*engine->texture)(engine->data, images + i);
		
		int count = 0;
		while (paragraph->text[count] != 0) count++;
		delays[i] = count * 4;
	}
	
	struct mimimi_chapter_I_toast_dialogue *toast_dialogue = mimimi_allocate(allocator, sizeof *toast_dialogue);
	toast_dialogue->count = dialogue->count;
	toast_dialogue->delays = delays;
	toast_dialogue->textures = textures;
	toast_dialogue->heights = heights;
	
	return toast_dialogue;
}

void mimimi_chapter_I_show_toast_dialogue_top(struct mimimi_behavior *behavior, struct mimimi_engine *engine, struct mimimi_chapter *chapter, struct mimimi_chapter_I_dialogue *dialogue, struct mimimi_allocator *allocator)
{
	struct mimimi_allocator *temporary_allocator = mimimi_temporary_allocator(allocator);
	struct mimimi_chapter_I_toast_dialogue *toast_dialogue = mimimi_chapter_I_toast_dialogue(dialogue, engine, temporary_allocator);
	
	for (int i = 0 ; i < toast_dialogue->count ; i++)
	{
		int height = toast_dialogue->heights[i];
		
		for (int j = 0 ; j < 16 ; j++)
		{
			(*engine->stamp)(engine->data, (128 + 8) * j / 16 - 128, 8, toast_dialogue->textures[i]);
			(*behavior->behave)(behavior->data);
		}
		
		for (int j = 0 ; j < toast_dialogue->delays[i] ; j++)
		{
			(*engine->stamp)(engine->data, 8, 8, toast_dialogue->textures[i]);
			(*behavior->behave)(behavior->data);
		}
		
		for (int j = 0 ; j < 32 ; j++)
		{
			(*engine->stamp)(engine->data, 8, 8 - height * j / 32, toast_dialogue->textures[i]);
			(*behavior->behave)(behavior->data);
		}
		
		(*engine->invalidate)(engine->data, toast_dialogue->textures[i]);
	}
	
	mimimi_finish_temporary_allocator(temporary_allocator);
}

void mimimi_chapter_I_show_toast_dialogue_bottom(struct mimimi_behavior *behavior, struct mimimi_engine *engine, struct mimimi_chapter *chapter, struct mimimi_chapter_I_dialogue *dialogue, struct mimimi_allocator *allocator)
{
	struct mimimi_allocator *temporary_allocator = mimimi_temporary_allocator(allocator);
	struct mimimi_chapter_I_toast_dialogue *toast_dialogue = mimimi_chapter_I_toast_dialogue(dialogue, engine, temporary_allocator);
	
	for (int i = 0 ; i < toast_dialogue->count ; i++)
	{
		int height = toast_dialogue->heights[i];
		
		for (int j = 0 ; j < 16 ; j++)
		{
			(*engine->stamp)(engine->data, 8, engine->size->height - j * height / 16, toast_dialogue->textures[i]);
			(*behavior->behave)(behavior->data);
		}
		
		for (int j = 0 ; j < toast_dialogue->delays[i] ; j++)
		{
			(*engine->stamp)(engine->data, 8, engine->size->height - height, toast_dialogue->textures[i]);
			(*behavior->behave)(behavior->data);
		}
		
		for (int j = 0 ; j < 32 ; j++)
		{
			(*engine->stamp)(engine->data, 8 - (128 + 8) * j / 32, engine->size->height - height, toast_dialogue->textures[i]);
			(*behavior->behave)(behavior->data);
		}
		
		(*engine->invalidate)(engine->data, toast_dialogue->textures[i]);
	}
	
	mimimi_finish_temporary_allocator(temporary_allocator);
}
